unit Unit1;

{$mode objfpc}{$H+}

// -----------------------------------------------------------------------------
//
// latest changes 2024-06-11
//
// adapted from https://forum.lazarus.freepascal.org/index.php/topic,35850.0.html
//
// -----------------------------------------------------------------------------


interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  StdCtrls, OpenGLContext, GL;



const
  FontCount = 100;


type
  { TForm1 }
  TForm1 = class(TForm)
    ButtonSelectFont: TButton;
    CheckBox1: TCheckBox;
    Edit1: TEdit;
    FontDialog1: TFontDialog;
    Image1: TImage;
    OpenGLControl1: TOpenGLControl;
    RadioGroup1: TRadioGroup;
    Timer1: TTimer;
    procedure ButtonSelectFontClick(Sender: TObject);
    procedure CheckBox1Change(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure OpenGLControl1Paint(Sender: TObject);
    procedure RadioGroup1SelectionChanged(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
  private
    FontData: array[0..127 - 32] of record
      TexturID, CharSize: GLuint;
      end;
    bit: TBitmap;

    SelectedFont: TFont;
    FontBitmap: TBitmap;
    FontSize: integer;
    FontCharWidth : array [0..FontCount-1] of integer;
    RectSize: Real;
    RectPosX: Real;
    RectPosY: Real;
    RectDirX: Real;
    RectDirY: Real;
    function CreateFontBitmap(F: TFont): TBitmap;
    procedure DisplayChar(C: Char);
    procedure DisplayString(S: String);
    function LoadFontBitmap: TBitmap;
  public
    { public declarations }
  end;


const
  size = 64;
const
  hello: PChar = 'Hello World !';


var
  Form1: TForm1;


implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
var
  i:integer;
begin
  FontSize:= 30;
  RectSize:= 1;
  RectPosX:= 0;
  RectPosY:= 0;
  RectDirX:= 0.02;
  RectDirY:= 0.02;

  // select one of the Bitmap Sources:
  //RadioGroup1.ItemIndex:= 0;
  RadioGroup1.ItemIndex:= 1;

  Image1.Width:= FontBitmap.Width;
  Image1.Height:= FontBitmap.Height;
  Image1.Picture.Bitmap.Assign( FontBitmap);

  // --- from project Texture_Font ---------------------------------------------
  OpenGLControl1.MakeCurrent;
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  for i := 0 to Length(FontData) - 1 do begin
    glGenTextures(1, @FontData[i].TexturID);
  end;
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glClearColor(0.8, 0.5, 0.3, 1.0);


  bit := TBitmap.Create;
  //  bit.PixelFormat := pf32bit;

  bit.Transparent := True;
  bit.Canvas.Brush.Style := bsClear;
  bit.Canvas.Font.Color := clYellow;
  bit.Canvas.Font.Height := 32;
  for i := 0 to Length(FontData) - 1 do begin
    bit.Clear;
    bit.SetSize(size, size);

    bit.Canvas.TextOut(0, 0, char(i + 32));
    FontData[i].CharSize := bit.Canvas.TextWidth(char(i + 32));

    glBindTexture(GL_TEXTURE_2D, FontData[i].TexturID);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    {$IFDEF Linux}
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, bit.Width, bit.Height, 0, GL_RGBA, GL_UNSIGNED_BYTE, bit.RawImage.Data);
    {$ENDIF}

    {$IFDEF Windows}
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, bit.Width, bit.Height, 0, GL_RGB, GL_UNSIGNED_BYTE, bit.RawImage.Data);
    {$ENDIF}
  end;
  bit.Free;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
  i: integer;
begin
  for i := 0 to Length(FontData) - 1 do begin
    glDeleteTextures(1, @FontData[i].TexturID);
  end;
end;

procedure TForm1.CheckBox1Change(Sender: TObject);
begin
  RadioGroup1SelectionChanged( Sender)
end;

procedure TForm1.ButtonSelectFontClick(Sender: TObject);
begin
  FontDialog1.Execute;

  FontBitmap.Free;
  FontBitmap:= CreateFontBitmap( FontDialog1.Font);
  Image1.Picture.Bitmap.Assign( FontBitmap);
end;

procedure TForm1.RadioGroup1SelectionChanged(Sender: TObject);
begin
  FontBitmap.Free; // don't forget this ..

  case RadioGroup1.ItemIndex of
  0: begin
       FontBitmap:= LoadFontBitmap;
       Image1.Picture.Bitmap.Assign( FontBitmap);
       CheckBox1.Visible:= FALSE;
       ButtonSelectFont.Visible:= FALSE;
       Edit1.Visible:= FALSE;
     end;
  1: begin
       FontBitmap:= CreateFontBitmap( SelectedFont);
       Image1.Picture.Bitmap.Assign( FontBitmap);
       CheckBox1.Visible:= TRUE;
       ButtonSelectFont.Visible:= TRUE;
       Edit1.Visible:= TRUE;
     end;
  end;
end;

procedure TForm1.OpenGLControl1Paint(Sender: TObject);
  procedure Draw(size: GLfloat);
  begin
    glBegin(GL_QUADS);
    glColor3f(1.0, 1.0, 1.0);
    glTexCoord2f(0.0, 1.0);
    glVertex3f(-size, -size, 0);
    glTexCoord2f(0.0, 0.0);
    glVertex3f(-size, size, 0);
    glTexCoord2f(1.0, 0.0);
    glVertex3f(size, size, 0);
    glTexCoord2f(1.0, 1.0);
    glVertex3f(size, -size, 0);
    glEnd();
  end;
const
  winkel: GLfloat = 0;
var
  i: integer;
  ofs: integer = 0;
begin
  // GL clear screen
  glLoadIdentity;
  glClearColor(0, 0, 0, 1);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  // GL preparation
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glEnable(GL_BLEND);
  glEnable(GL_DEPTH_TEST);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


  // ---------------------
  glEnable(GL_TEXTURE_2D);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

  // To define texture images, call glTexImage2D:
  glTexImage2D( GL_TEXTURE_2D, 0, 4, FontBitmap.Width, FontBitmap.Height, 0,
                GL_RGB, GL_UNSIGNED_BYTE, FontBitmap.RawImage.Data);


  // Draw the text
  glTranslatef( -0.9, 0.8, 0);
  DisplayString( 'X' +IntToStr( Mouse.CursorPos.X));
  glTranslatef( 0.2, 0, 0);
  DisplayString( 'Y' +IntToStr( Mouse.CursorPos.Y));
  // After finished rendering a texture, you need to disable a texture:
  glDisable( GL_TEXTURE_2D);


  // Draw the rectangle
  glLoadIdentity;
  glTranslatef( RectPosX, RectPosY, 0.1);
  glScalef( RectSize, RectSize, RectSize);
  glBegin(GL_QUADS);
  glColor3ub(0, 255, 0);
  glVertex3f(-0.5, -0.5, 0);
  glVertex3f(0.5, -0.5, 0);
  glVertex3f(0.5, 0.5, 0);
  glVertex3f(-0.5, 0.5, 0);
  glEnd();


  // ---  draw HELLO WOLRD ---
  winkel += 1.1;

  //glMatrixMode(GL_MODELVIEW);  ??
  glLoadIdentity();

  glEnable(GL_TEXTURE_2D);
  for i := 0 to Length(hello) - 1 do begin
    glPushMatrix;
    glRotatef(winkel, 0.0, 0.0, 1.0);
    glBindTexture(GL_TEXTURE_2D, FontData[byte(hello[i]) - 32].TexturID);
    glTranslatef(ofs / 80 - 0.3, 0.0, 0.0);
    {%H-}Inc(ofs, FontData[byte(hello[i]) - 32].CharSize);
    Draw(0.4);
    glPopMatrix;
  end;
  // After finished rendering a texture, you need to disable a texture:
  glDisable( GL_TEXTURE_2D);


  OpenGLControl1.SwapBuffers;
end;


procedure TForm1.Timer1Timer(Sender: TObject);
var
  R: Real;
begin
  R := Random/20 - 1/40;
  RectSize := RectSize + R;
  if (RectSize > 1) then RectSize := 1;
  if (RectSize < 0.1) then RectSize := 0.1;

  RectPosX := RectPosX + RectDirX;
  RectPosY := RectPosY + RectDirY;
  if (RectPosX > 0.8) or (RectPosX < -0.8) then RectDirX := -RectDirX;
  if (RectPosY > 0.8) or (RectPosY < -0.8) then RectDirY := -RectDirY;
  if (Random < 0.1) then RectDirX := -RectDirX;
  if (Random < 0.1) then RectDirY := -RectDirY;


  OpenGLControl1.Invalidate;  // OpenGL does not "repaint" without Invalidate() here
end;


const
  CharWidth  = 0.13;
  CharHeight = 0.17;

procedure TForm1.DisplayChar(C: Char);
var
  FontPos    : Byte;
  FontWidth  : real;
  FontHeight : real;
  X, Y       : real;
begin
  FontPos := 0;

  case RadioGroup1.ItemIndex of
  0: begin
       // missing here:  (C >= 'a') and (C <= 'z') ..
       if (C >= 'A') and (C <= 'Z') then FontPos := ord(C) - 38;
       if (C >= '0') and (C <= '9') then FontPos := ord(C) + 5;
     end;
  1: begin
       // missing here:  (C >= 'a') and (C <= 'z') ..
       if (C >= 'A') and (C <= 'Z') then FontPos := ord(C) - 32;
       if (C >= '0') and (C <= '9') then FontPos := ord(C) - 32;
     end;
  end;
  if (FontPos = 0) then Exit;

  FontWidth  := FontSize / (FontBitmap.Width);  // FontSize is the width of the font
  FontHeight := FontSize / (FontBitmap.Height); // FontSize is the height of the font
  X := ((FontPos -1) mod 10) * FontWidth;
  Y := ((FontPos -1) div 10) * FontHeight;

  glBegin( GL_QUADS);
  glColor3ub( 255, 255, 255);
  glTexCoord2f( X, Y +FontHeight);
  glVertex3f( 0, 0, 0);
  glTexCoord2f( X +FontWidth, Y +FontHeight);
  glVertex3f( CharWidth, 0, 0);
  glTexCoord2f( X +FontWidth, Y);
  glVertex3f( CharWidth, CharHeight, 0);
  glTexCoord2f( X, Y);
  glVertex3f( 0, CharHeight, 0);
  glEnd();
end;

procedure TForm1.DisplayString(S: String);
var
  i: Integer;
begin
  for i := 1 to Length(S) do begin
    DisplayChar( S[i]);

    // Die glTranslatef-Funktion multipliziert die aktuelle Matrix mit einer Übersetzungsmatrix.
    // https://learn.microsoft.com/de-de/windows/win32/opengl/gltranslatef
    glTranslatef( CharWidth, 0, 0);
  end;
end;

function TForm1.LoadFontBitmap: TBitmap;
const
  FontBitmapFile = 'FontTexture.bmp';   // chars of 30 x 30 pixel
begin
  result:= TBitmap.Create;
  result.LoadFromFile( FontBitmapFile);
end;

// from https://www.lazarusforum.de/viewtopic.php?p=141767#p141767
function TForm1.CreateFontBitmap(F: TFont): TBitmap;
var
  i : integer;
  x : integer;
begin
  Result := TBitmap.Create;
  Result.Width := (FontSize) * 10;
  Result.Height := (FontSize) * 10;

  if CheckBox1.Checked
    then Result.Canvas.FloodFill( 0, 0, clWhite, fsBorder);

  Result.Canvas.Font.Color  := clRed;
  //Result.Canvas.Font.Color  := clGreen;
  //Result.Canvas.Font.Color  := clBlack;
  if F = TFont(NIL)
    then Result.Canvas.Font.Name:= 'monospace'
    else Result.Canvas.Font:= F;
  Edit1.Text:= Result.Canvas.Font.Name;
  Edit1.Font:= Result.Canvas.Font;
  Result.Canvas.Font.Height := FontSize;

  for i := 0 to FontCount - 1 do begin
    // TextOut() schreibt einen String beginnend beim Punkt (X,Y) auf die Zeichenfläche
    // und setzt die neue Stiftposition auf das Ende des Strings.
    x:= (i mod 10) * FontSize +2;
    Result.Canvas.TextOut( x, (i div 10) * FontSize +1, char(i +33));

    FontCharWidth[i]:= Result.Canvas.PenPos.X -x;
  end;
end;


end.
